from csv import DictReader
from enum import Enum
from typing import Callable, List, Union

from .parsers.engine import parser as engine_parser
from .parsers.rekordbox import parser as rekordbox_parser
from .parsers.serato import parser as serato_parser
from .parsers.traktor import parser as traktor_parser
from .parsers.virtualdj import parser as virtualdj_parser
from .track import Track


class PlaylistType(Enum):
    UNKNOWN = 0
    ENGINE = 1
    REKORDBOX = 2
    SERATO = 3
    TRAKTOR = 4
    VIRTUALDJ = 5


class PlaylistParser(object):
    def __init__(
        self,
        file_path: str,
        *,
        require_title: bool = True,
        require_duration: bool = False,
        require_year: bool = False,
        require_bpm: bool = False,
        require_fp: bool = False,
        default_artist: str = "Unknown Artist",
        verbose: bool = False,
    ):
        self.file_path = file_path
        self.verbose = verbose
        self.playlist_type: PlaylistType = PlaylistType.UNKNOWN
        self.tracks: List[Track] = []
        self._parser: Union[Callable[[str], List], None] = None
        self.is_parsed = False
        self.require_title = require_title
        self.require_duration = require_duration
        self.require_bpm = require_bpm
        self.require_year = require_year
        self.require_fp = require_fp
        self.default_artist = default_artist

        self._determine_type()

    def _determine_type(self):
        if self.file_path.endswith(".nml"):
            if self.verbose:  # pragma: no cover
                print("Looking for nml")
            self._parser = traktor_parser
            self.playlist_type = PlaylistType.TRAKTOR
        elif self.file_path.endswith(".txt"):
            if self.verbose:  # pragma: no cover
                print("Looking for txt")
            self._parser = rekordbox_parser
            self.playlist_type = PlaylistType.REKORDBOX
        elif self.file_path.endswith(".csv"):
            if self.verbose:  # pragma: no cover
                print("Looking for csv. filepath: ", self.file_path)

            with open(self.file_path) as file:
                reader = DictReader(file)
                line = reader.__next__()
                if self.verbose:  # pragma: no cover
                    print("First line: ", line)

                if "\ufeffsep=" in line.keys():
                    self._parser = virtualdj_parser
                    self.playlist_type = PlaylistType.VIRTUALDJ

                if "#" in line.keys():
                    self._parser = engine_parser
                    self.playlist_type = PlaylistType.ENGINE

                if "name" in line.keys():
                    self._parser = serato_parser
                    self.playlist_type = PlaylistType.SERATO

        if self.verbose:  # pragma: no cover
            print(f"Initialized {self.playlist_type} playlist.")

        if self.playlist_type == PlaylistType.UNKNOWN:
            raise Exception(f"Couldn't determine playlist type from '{self.file_path}'")

    def parse(self):
        if self.verbose:  # pragma: no cover
            print("Parsing...")
        self.tracks = self._parser(
            self.file_path,
            require_title=self.require_title,  # type: ignore
            require_duration=self.require_duration,  # type: ignore
            require_bpm=self.require_bpm,  # type: ignore
            require_year=self.require_year,  # type: ignore
            require_fp=self.require_fp,  # type: ignore
            default_artist=self.default_artist,  # type: ignore
            verbose=self.verbose,  # type: ignore
        )  # type: ignore
        self.is_parsed = True
        if self.verbose:  # pragma: no cover
            print(f"Found {len(self.tracks)} tracks.")

    def get_tracks(self):
        if not self.is_parsed:
            self.parse()
        return self.tracks


__version__ = "3.0.0-beta.8"



def time_str_to_seconds(time: str) -> int:
    """
    Convert a time string to seconds.

    0:01:00 -> 60
    """
    try:
        secs = sum(x * int(t) for x, t in zip([1, 60, 3600], reversed(time.split(":"))))
    except Exception:
        secs = 0
    return secs



import xml.etree.ElementTree as ET

from ..track import Track


def parser(
    file_path,
    *,
    require_title=True,
    require_duration=False,
    require_year=False,
    require_bpm=False,
    require_fp=False,
    default_artist="",
    verbose=False,
):
    """
    Traktor supports:
    - title
    - artist
    - year
    - playtime
    - bpm
    """
    if require_fp:
        raise NotImplementedError("Traktor parser doesn't support file paths.")

    tracks = []
    traktor_xml = ""
    counter = 0

    with open(file_path) as file:
        traktor_xml = ET.fromstring(file.read())

    entries = traktor_xml.findall("COLLECTION/ENTRY")

    for track in entries:
        playtime = 0
        year = ""
        bpm = 0

        try:
            track_title = track.get("TITLE", "").strip()
            track_artist = track.get("ARTIST", "").strip()
            if not track_artist:
                track_artist = default_artist

            meta = track.find("INFO")
            if meta is not None:
                playtime = int(meta.get("PLAYTIME", ""))
                # key = meta.get("KEY", "")
                year = meta.get("RELEASE_DATE", "")
                if verbose:  # pragma: no cover
                    print(f"found year: {year}, playtime: {playtime}")

            tempometa = track.find("TEMPO")
            if tempometa is not None:
                bpm = int(float(tempometa.get("BPM", 0)))

            tracks.append(
                Track(
                    title=track_title,
                    artist=track_artist,
                    year=year,
                    duration=playtime,
                    bpm=bpm,
                )
            )
        except Exception as e:  # pragma: no cover
            print(f"Skipping line {counter}", e)

        counter += 1

    return tracks



from csv import DictReader

from ..track import Track


def parser(
    file_path,
    *,
    require_title=True,
    require_duration=False,
    require_year=False,
    require_bpm=False,
    require_fp=False,
    default_artist="",
    verbose=False,
):
    """
    Serato supports:
    - title
    - artist
    - year
    """
    if require_duration:
        raise NotImplementedError("Serato parser doesn't support require_duration.")

    if require_year:
        raise NotImplementedError("Serato parser doesn't support require_year.")

    if require_bpm:
        raise NotImplementedError("Serato parser doesn't support require_bpm.")

    if require_fp:
        raise NotImplementedError("Serato parser doesn't support file paths.")

    with open(file_path) as file:
        reader = DictReader(file)
        tracks = []
        counter = 0

        for line in reader:
            if counter > 0:
                try:
                    title = line["name"].strip()
                    artist = line["artist"].strip()
                    if not artist:
                        artist = default_artist
                    year = line["year"].strip()
                    tracks.append(Track(title=title, artist=artist, year=year))
                except Exception as e:  # pragma: no cover
                    print(f"Skipping line {counter}", e)
            counter += 1
        return tracks



from datetime import timedelta
from typing import Dict, Union
from unicodedata import normalize


class Track(object):
    def __init__(
        self,
        *,
        title: str,
        artist: str,
        duration: int = 0,
        year: str = "",
        bpm: int = 0,
        file_path: str = "",
    ):
        self.title = normalize("NFC", title).strip()
        self.artist = normalize("NFC", artist).strip()
        self.duration = duration
        self.year = self._clean_year(year)
        self.bpm = bpm
        self.file_path = file_path

    def __str__(self):
        return f"{self.artist} - {self.title}"

    def _clean_year(self, year: str):
        cleaned_year = year
        if "-" in year:
            cleaned_year = year.split("-")[0]
        elif "/" in year:
            cleaned_year = year.split("/")[0]
        else:
            cleaned_year = year
        try:
            return int(cleaned_year)
        except Exception:
            return 0

    def duration_str(self):
        if self.duration:
            string = str(timedelta(seconds=self.duration))

            # 0:00:12
            if string.startswith("0:"):
                string = string[2:]

            # 00:12
            if string.startswith("00:"):
                string = "0:" + string[3:]

            # 01:12
            if string.startswith("0") and not string.startswith("0:"):
                string = string[1:]
            return string
        else:
            return "0:00"

    def as_dict(self, no_meta=False):
        song: Dict[str, Union[str, int]] = {
            "title": self.title,
            "artist": self.artist,
        }
        if no_meta:
            return song
        if self.duration:
            song["duration"] = self.duration
            song["duration_str"] = self.duration_str()
        if self.year:
            song["year"] = self.year
        if self.bpm:
            song["bpm"] = self.bpm
        if self.file_path:
            song["file_path"] = self.file_path
        return song



import codecs
import csv
from io import StringIO

from ..track import Track
from ..utils import time_str_to_seconds


def parser(
    file_path,
    *,
    require_title=True,
    require_duration=False,
    require_year=False,
    require_bpm=False,
    require_fp=False,
    default_artist="",
    verbose=False,
):
    """
    Rekordbox supports:
    - title
    - artist
    - year
    - playtime
    - bpm
    - file_path
    """
    with open(file_path, "rb") as handle, codecs.EncodedFile(
        handle, data_encoding="utf-8", file_encoding="utf-16", errors="ignore"
    ) as stream:

        reader = csv.DictReader(StringIO(stream.read().decode("utf-8")), delimiter="\t")  # type: ignore

        tracks = []
        counter = 0

        for line in reader:
            artist = ""
            title = "Unknown"
            playtime = 0
            year = ""
            bpm = 0
            file_path = ""

            try:
                title = line["Track Title"].strip()
            except KeyError:
                if require_title:
                    raise ValueError("Title required but not found in file.")

            try:
                playtime = time_str_to_seconds(line["Time"].strip())
            except KeyError:
                if require_duration:
                    raise ValueError("Duration required but not found in file.")

            try:
                bpm = int(float(line["BPM"].strip()))
            except KeyError:
                if require_bpm:
                    raise ValueError("BPM required but not found in file.")

            try:
                year = line["Year"].strip()
            except KeyError:
                if require_year:
                    raise ValueError("Year required but not found in file.")

            try:
                file_path = line["Location"].strip()
            except Exception:
                if require_fp:
                    raise ValueError("File paths required but not found in file.")

            artist = line["Artist"].strip()
            if not artist:
                artist = default_artist

            tracks.append(
                Track(
                    title=title,
                    artist=artist,
                    year=year,
                    duration=playtime,
                    bpm=bpm,
                    file_path=file_path,
                )
            )
            counter += 1
        return tracks



import csv

from ..track import Track
from ..utils import time_str_to_seconds


def parser(
    file_path,
    *,
    require_title=True,
    require_duration=False,
    require_year=False,
    require_bpm=False,
    require_fp=False,
    default_artist="",
    verbose=False,
):
    """
    VirtualDJ supports:
    - title
    - artist
    - year
    - playtime
    - bpm
    """
    if require_duration:
        raise NotImplementedError("VirtualDJ parser doesn't support require_duration.")

    if require_year:
        raise NotImplementedError("VirtualDJ parser doesn't support require_year.")

    if require_bpm:
        raise NotImplementedError("VirtualDJ parser doesn't support require_bpm.")

    if require_fp:
        raise NotImplementedError("VirtualDJ parser doesn't support file paths.")

    with open(file_path) as file:
        reader = csv.DictReader(
            file, fieldnames=["Title", "Artist", "Remix", "Length", "Bpm", "Key", "Year"]
        )
        tracks = []
        counter = 0

        for line in reader:
            if counter > 1:
                try:
                    title = line["Title"].strip()
                    artist = line["Artist"].strip()
                    if not artist:
                        artist = default_artist

                    playtime = time_str_to_seconds(line["Length"].strip())
                    bpm = int(float(line["Bpm"].strip()))
                    try:
                        year = line["Year"].strip()
                    except KeyError:
                        year = ""
                    tracks.append(
                        Track(title=title, artist=artist, year=year, duration=playtime, bpm=bpm)
                    )
                except Exception as e:  # pragma: no cover
                    if verbose:
                        print(f"Line {counter}: ", line)
                    print(f"Skipping line {counter}", e)

            counter += 1
        return tracks



import csv

from ..track import Track


def parser(
    file_path,
    *,
    require_title=True,
    require_duration=False,
    require_year=False,
    require_bpm=False,
    require_fp=False,
    default_artist="",
    verbose=False,
):
    """
    Engine supports:
    - title
    - artist
    - year
    - playtime
    - bpm
    - bpm
    - file_path
    """
    with open(file_path) as file:
        reader = csv.DictReader(file)
        tracks = []
        counter = 0

        for line in reader:
            try:
                title = line["Title"].strip()
                artist = line["Artist"].strip()
                if not artist:
                    artist = default_artist
                year = line["Year"].strip()
                bpm = int(line["BPM"].strip())
                playtime = int(line["Length"].strip())
                file_path = line["File name"].strip()
                tracks.append(
                    Track(
                        title=title,
                        artist=artist,
                        year=year,
                        duration=playtime,
                        bpm=bpm,
                        file_path=file_path,
                    )
                )
            except Exception as e:  # pragma: no cover
                print(f"Skipping line {counter}", e)

            counter += 1
        return tracks
